@@ -2,6 +2,7 @@
 
 import sys
 import csv
+import sqlite3
 import pandas as pd
 import numpy as np
 from warnings import warn
@@ -19,18 +20,12 @@ from typing import Union, Any, List, Tuple, FrozenSet
 class SparseTensorClassifier:
     """
     This class implements the Sparse Tensor Classifier (STC), a supervised classification algorithm for categorical
-    data inspired by the notion of superposition of states in quantum physics. It supports multiclass and multilabel
-    classification, online learning, prior knowledge, automatic dataset balancing, missing data, and provides a native
-    explanation of its predictions both for single instances and for each target class label globally.
+    data inspired by the notion of superposition of states in quantum physics.
 
     The algorithm is implemented in SQL. By default, the library uses an in-memory SQLite database, shipped with
-    Python standard library, that require no configuration by the user. However, it is possible to configure STC to
-    run on alternative DBMS in order to take advantage of persistent storage and scalability. Supported DBMS are:
-    **SQLite** v3.24+;
-    **MySQL** v8.0+;
-    **PostgreSQL** 9.6+;
-    **SQL Server** 2017+;
-    **Oracle** 12.2+.
+    Python standard library, that require no configuration by the user. It is also possible to configure STC to
+    run on `alternative DBMS <https://github.com/SparseTensorClassifier/tutorial/blob/main/Quickstart_DBMS.ipynb>`_
+    in order to take advantage of persistent storage and scalability.
 
     The input data must be a pandas ``DataFrame`` or a JSON structured as follows:
 
@@ -58,7 +53,7 @@ class SparseTensorClassifier:
                      with :meth:`stc.SparseTensorClassifier.learn`
     :param engine: Connection to database in the form of a
                    `SQLAlchemy engine <https://docs.sqlalchemy.org/en/13/core/engines.html>`_.
-                   By default, STC uses an in-memory SQLite database. The list of supported DBMS is provided above.
+                   By default, STC uses an in-memory SQLite database.
     :param prefix: Prefix to use in the database tables. STC instances initialized with different prefix are completely
                    independent. This makes possible to use the same ``engine`` multiple times with different ``prefix``,
                    without creating a new database/schema. If an instance associated with the same ``engine`` and
@@ -80,13 +75,15 @@ class SparseTensorClassifier:
                  or ``log`` for cross-entropy (log-loss).
     :param tol: The actual predicted probabilities are replaced with the value of ``tol`` when using ``loss='log'`` and
                 the actual predicted probability is zero.
+    :param verbose: Print on progress? Default ``True``.
     """
 
     def __init__(self, targets: List[str], features: List[str] = None, collapse: bool = True,
                  engine: Union[Engine, str] = "sqlite://", prefix: str = "stc",
                  chunksize: int = 100, cache: bool = True,
                  power: float = 0.5, balance: float = 1, entropy: float = 1,
-                 loss: str = 'norm', tol: float = 1e-15) -> None:
+                 loss: str = 'norm', tol: float = 1e-15,
+                 verbose: bool = True) -> None:
 
         # validate
         if not targets:
@@ -166,15 +163,11 @@ class SparseTensorClassifier:
                 raise Exception("Dimensions found in DB different from the dimensions provided")
 
         # init values map from DB or from scratch
-        self.values_map = {}
         try:
             x = self.read_sql(f"SELECT * FROM {self.value_table}")
-            for d in x[self.dim_field].unique():
-                v = x[x[self.dim_field] == d]
-                self.values_map[d] = self._defaultdict(zip(v[self.name_field], v[self.value_field]))
+            self.values_map = self._defaultdict(zip(x[self.name_field], x[self.value_field]))
         except SQLAlchemyError:
-            for d in self.dims_map.values():
-                self.values_map[d] = self._defaultdict()
+            self.values_map = self._defaultdict()
 
         # encrypt
         self.dims = self._encrypt(self.dims)
@@ -188,6 +181,7 @@ class SparseTensorClassifier:
         self.entropy = entropy
         self.loss = loss
         self.tol = tol
+        self.verbose = verbose
         self.qtable = {}
         self.set({})
 
@@ -336,9 +330,6 @@ class SparseTensorClassifier:
             items = items_all[c:c + self.chunksize]
             uids = uids_all[c:c + self.chunksize]
 
-            # tmp dictionary
-            values_map = self.values_map.copy()
-
             # add key:value pairs to the dictionary
             new = []
             for i, item in enumerate(items):
@@ -350,12 +341,11 @@ class SparseTensorClassifier:
 
                 for dim, values in dv:
                     try:
-                        k = self.dims_map[dim]
                         for v in set(values):
-                            n = len(values_map[k])
-                            m = values_map[k].setdefault(v, n)
+                            n = len(self.values_map)
+                            m = self.values_map.setdefault(v, n)
                             if n == m:
-                                new.append({self.dim_field: k, self.name_field: v, self.value_field: m})
+                                new.append({self.name_field: v, self.value_field: m})
                     except KeyError:
                         pass
 
@@ -363,7 +353,6 @@ class SparseTensorClassifier:
             if new:
                 x = pd.DataFrame(new)
                 self.to_sql(x, self.value_table, if_exists="append")
-                self.values_map = values_map
 
             # transform
             self._transform(items=items, uids=uids, dims=self.dims,
@@ -374,7 +363,8 @@ class SparseTensorClassifier:
             if_exists = "append"
 
             # update progressbar
-            progress.update(c + self.chunksize)
+            if self.verbose:
+                progress.update(c + self.chunksize)
 
     def explain(self, features: List[str] = None) -> pd.DataFrame:
         """
@@ -506,7 +496,8 @@ class SparseTensorClassifier:
             self._DROP(tmp_test)
 
             # update progressbar
-            progress.update(c + self.chunksize)
+            if self.verbose:
+                progress.update(c + self.chunksize)
 
         # back to original dims and values
         t = self._decrypt(self.targets)
@@ -662,7 +653,8 @@ class SparseTensorClassifier:
                         self.qtable[key] = val
 
                     # update progressbar
-                    progress.step()
+                    if self.verbose:
+                        progress.step()
 
                     # check runtime
                     toc = time()
@@ -798,7 +790,7 @@ class SparseTensorClassifier:
         for d in x.columns:
             if d in self.dims_map:
                 i = self.dims_map[d]
-                y['d' + str(i)] = x[d].map(self.values_map[i])
+                y['d' + str(i)] = x[d].map(self.values_map)
             else:
                 y[d] = x[d]
 
@@ -812,15 +804,13 @@ class SparseTensorClassifier:
         if isinstance(x, list):
             return [dinv[d] for d in x]
 
-        vinv = {}
-        for d, m in self.values_map.items():
-            vinv['d' + str(d)] = {v: k for k, v in m.items()}
+        vinv = {v: k for k, v in self.values_map.items()}
 
         y = pd.DataFrame()
         for i in x.columns:
             if i in dinv:
                 d = dinv[i]
-                y[d] = x[i].map(vinv[i])
+                y[d] = x[i].map(vinv)
             else:
                 y[i] = x[i]
 
@@ -1172,7 +1162,7 @@ class SparseTensorClassifier:
                             m = 0
                             k = self.dims_map[dim]
                             for v, n in Counter(value_list).most_common():
-                                v = self.values_map[k][v]
+                                v = self.values_map[v]
                                 if v == -1:
                                     m += n
                                 else:
@@ -1439,9 +1429,8 @@ class SparseTensorClassifier:
 
         if self.db == "sqlite":
             sql = (
-                f"INSERT INTO {self.meta_table} ({self.name_field}, {self.value_field}) "
-                f"VALUES ('{key}', :value) "
-                f"ON CONFLICT ({self.name_field}) DO UPDATE SET {self.value_field}=excluded.{self.value_field}")
+                f"INSERT OR REPLACE INTO {self.meta_table} ({self.name_field}, {self.value_field}) "
+                f"VALUES ('{key}', :value)")
         elif self.db == "postgresql":
             sql = (
                 f"INSERT INTO {self.meta_table} ({self.name_field}, {self.value_field}) "
@@ -1499,12 +1488,19 @@ class SparseTensorClassifier:
             data = self._SELECT_data(dims=dims, items_table=items_table, corpus=True)
 
         if self.db == "sqlite":
-            sql = (
-                f"WITH corpus_ AS ({data}) "
-                f"INSERT INTO {corpus_table} ({d}, {s}) "
-                f"SELECT * FROM corpus_ WHERE true "
-                f"ON CONFLICT ({d}) "
-                f"DO UPDATE SET {s}={corpus_table}.{s}+excluded.{s}")
+            if sqlite3.sqlite_version_info < (3, 24, 0):
+                sql = (
+                    f"WITH corpus_ AS ({data}) "
+                    f"INSERT OR REPLACE INTO {corpus_table} ({d}, {s}) "
+                    f"SELECT {self._COLS('corpus_', dims)}, corpus_.{s}+COALESCE({corpus_table}.{s}, 0) AS {s} "
+                    f"FROM corpus_ LEFT JOIN {corpus_table} {self._ON_NATURAL('corpus_', corpus_table, dims)}")
+            else:
+                sql = (
+                    f"WITH corpus_ AS ({data}) "
+                    f"INSERT INTO {corpus_table} ({d}, {s}) "
+                    f"SELECT * FROM corpus_ WHERE true "
+                    f"ON CONFLICT ({d}) "
+                    f"DO UPDATE SET {s}={corpus_table}.{s}+excluded.{s}")
         elif self.db == "postgresql":
             sql = (
                 f"WITH corpus_ AS ({data}) "
